#!/bin/sh

# dpi.sh
# version 1.3 (2014-03-03)
# by Adam Bjornson

# Description
# Returns the physical dimensions, supported resolutions, aspect ratio, and/or DPI of any connected display. Parses the low-level EDID of the specified display device. Designed for Mac OS X (using ioreg) but may work with other systems

# See Also
# http://en.wikipedia.org/wiki/Extended_display_identification_data
# http://whitehatty.wordpress.com/2012/04/28/

# ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
# GENERAL
# ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

USAGE="\n\
dpi.sh \n\
Returns the physical dimensions, supported resolutions, aspect ratio, and/or DPI of any connected display. \n\n\
Usage: "$(basename "$0")" -d <number> -h -n <number> -q -r -s -v \n\
-d <number> - return screen dpi for the specified horizontal resolution \n\
-h - return horizontal screen size in inches \n\
-n <number> - specify the screen number (default 0, or main) \n\
-r - list supported screen resolutions, ratios, and dpis \n\
-q - output raw values \n\
-s - return diagonal screen size in inches \n\
-v - return vertical screen size in inches \n";

# If no arguments, echo usage and exit
[ -z "$1" ] && echo $USAGE && exit 1;

# Set defaults
FLAG_D=0;
FLAG_H=0;
FLAG_N=0;
FLAG_Q=0;
FLAG_R=0;
FLAG_S=0;
FLAG_V=0;

# Get arguments
OPTARG=""; OPTIND=1; OPTERR=0;
while getopts ":d:hn:qrsv" VALUE "$@" ; do
	[ "$VALUE" = "d" ] && FLAG_D="$OPTARG";
	[ "$VALUE" = "h" ] && FLAG_H=1;
	[ "$VALUE" = "n" ] && FLAG_N="$OPTARG";
	[ "$VALUE" = "q" ] && FLAG_Q=1;
	[ "$VALUE" = "r" ] && FLAG_R=1;
	[ "$VALUE" = "s" ] && FLAG_S=1;
	[ "$VALUE" = "v" ] && FLAG_V=1;
    if [ "$VALUE" = ":" ] ; then
		echo "Flag -$OPTARG requires an argument." && echo $USAGE && exit 1;
    fi
    if [ "$VALUE" = "?" ] ; then
		echo "Unknown flag -$OPTARG." && echo $USAGE && exit 1;
    fi
done

# ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
# HELPER FUNCTIONS
# ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

# Return stdin as $STDIN for use inside functions
GET_STDIN() { [ ! -t 0 ] && read -r STDIN; }

# Get a byte pair (00-FF) from a hex string, zero-based.
# get_byte hex_string byte_index
get_byte() {
	START=$(echo '('$2'*2)+1' | bc);
	END=$(echo $START'+1' | bc);
	echo $(echo $1 | cut -c$START-$END);
}

# Convert a hex number to a decimal or binary number
# Arguments: number (from standard input)
# Note: binary numbers are zero-padded to eight digits
# Note: bc requires capital hex letters (A-F, not a-f)
hex_to_dec() { GET_STDIN; echo $(printf '%d\n' '0x'$STDIN); }
hex_to_bin() { GET_STDIN; echo $(echo 'obase=2;ibase=16;'$STDIN | BC_LINE_LENGTH=9999 bc | awk '{ printf "%08d\n", $0 }'); }

# Return a sequence of bits from a binary number string
# Arguments: string, from, to
bin_to_bits() { echo $(echo $1 | cut "-c$2-$3"); }

# Convert an EDID-format horizontal resolution to its original value
# Arguments: x_res (from standard input)
edid_res() { GET_STDIN; echo $(echo '(('$STDIN'*8)+248)' | bc); }

# Convert an EDID-format aspect ratio string to its original value
# Arguments: 8-bit binary string (from standard input)
edid_aspect() {
	GET_STDIN;
	ASPECT=$(echo $STDIN | cut '-c1-2');
	case $ASPECT in
		( '00' ) echo '16:10' ;;
		( '01' ) echo '4:3' ;;
		( '10' ) echo '5:4' ;;
		( '11' ) echo '16:9' ;;
		( * ) echo $ASPECT ;;
	esac
}

# Convert an EDID-format aspect ratio scalar to its original value
# Arguments: 8-bit binary string (from standard input)
edid_scale() {
	GET_STDIN;
	SCALE=$(echo $STDIN | cut '-c1-2');
	case $SCALE in
		( '00' ) echo '0.625' ;;
		( '01' ) echo '0.75' ;;
		( '10' ) echo '0.8' ;;
		( '11' ) echo '0.5625' ;;
		( * ) echo $SCALE ;;
	esac
}

# ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––
# PARSE IOREG AND EDID
# ––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

# Get ioreg output
# (like '"IODisplayEDID" = <00ff...>')
IOREG=$(ioreg -lw0 -r -n 'display'$FLAG_N -d 2 | grep 'IODisplayEDID');

# Trim ioreg output
# (remove everything except EDID hex string, like '00ff...')
# (also make sure the hex digits are uppercase, like A-F)
HEX=$(echo $IOREG | sed "/[^<]*</s///" | sed "/[>*]/s///" | tr '[:lower:]' '[:upper:]');

# Get physical display size in inches (horizontal, vertical, diagonal)
# Take chars 43-44 (byte 21), convert to decimal, then from cm to in
H_SIZE=$(echo $HEX | cut -c43-44);
H_SIZE=$(printf '%d\n' '0x'$H_SIZE);
H_SIZE=$(echo $H_SIZE'*0.393701' | bc);
V_SIZE=$(echo $HEX | cut -c45-46);
V_SIZE=$(printf '%d\n' '0x'$V_SIZE);
V_SIZE=$(echo $V_SIZE'*0.393701' | bc);
D_SIZE=$(echo 'sqrt(('$H_SIZE'^2)+('$V_SIZE'^2))' | bc);

# Debug
# DEBUG_123=$(get_byte $HEX 38); # (byte 38 = chars 77-78) ok
# DEBUG_123=$(edid_x_res $(hex_to_dec $(get_byte $HEX 38))); # ok
# DEBUG_123=$(get_byte $HEX 38 | hex_to_dec | edid_x_res); # ok
# echo "DEBUG_123 = $DEBUG_123"; # ok

# Notes
# - Vertical resolution is based on horizontal resolution and aspect ratio
# - Four preferred resolution values are provided in the basic EDID data
# - The values are not necessarily in order of preference or size
# - Bytes 38-53 are 8x 2-byte fields of standard timing info
#   8-bit x resolution ((res*8(+248)) (256-2288px),
#   2-bit ratio, 6-bit vertical frequency (freq+60; 60-123 Hz)

# Physical Dimensions
# echo "Physical Dimensions = $H_SIZE\" x $V_SIZE\" ($D_SIZE\" diagonal)";

# Horizontal, Vertical, and Diagonal
if [ ! $FLAG_H = 0 ] ; then
	[ $FLAG_Q = 0 ] && echo "Horizontal = $H_SIZE\"" || echo "$H_SIZE";
fi
if [ ! $FLAG_V = 0 ] ; then
	[ $FLAG_Q = 0 ] && echo "Vertical = $V_SIZE\"" || echo "$V_SIZE";
fi
if [ ! $FLAG_S = 0 ] ; then
	[ $FLAG_Q = 0 ] && echo "Diagonal = $D_SIZE\"" || echo "$D_SIZE";
fi

# DPI
if [ ! $FLAG_D = 0 ] ; then
	X_RES=$FLAG_D;
	DPI=$(echo $X_RES'/'$H_SIZE | bc);
	[ $FLAG_Q = 0 ] && echo "DPI @ $FLAG_DPI = $DPI" || echo "$DPI";
fi

# Supported Resolutions, Ratios, and DPI
get_resolution () {
	BYTE_1=$(echo '38+(('$1'-1)*2)' | bc);
	BYTE_2=$(echo $BYTE_1'+1' | bc);
	X_RES=$(get_byte $HEX $BYTE_1 | hex_to_dec | edid_res);
	ASPECT=$(get_byte $HEX $BYTE_2 | hex_to_bin | edid_aspect);
	SCALE=$(get_byte $HEX $BYTE_2 | hex_to_bin | edid_scale);
	Y_RES=$(echo 'scale=0;('$X_RES'*'$SCALE')/1' | bc);
	DPI=$(echo $X_RES'/'$H_SIZE | bc);
	if [ $FLAG_Q = 0 ] ; then
		echo "Resolution $1 = $X_RES x $Y_RES ($ASPECT, $DPI DPI)";
	else
		echo "$X_RES $Y_RES $ASPECT $DPI";
	fi
}
if [ ! $FLAG_R = 0 ] ; then
	get_resolution 1;
	get_resolution 2;
	get_resolution 3;
	get_resolution 4;
fi

# Finish up
exit 0;